@@ -5,11 +5,11 @@
 ;; Author: Daniel Mendler <mail@daniel-mendler.de>
 ;; Maintainer: Daniel Mendler <mail@daniel-mendler.de>
 ;; Created: 2022
-;; Version: 0.2
+;; Version: 0.3
 ;; Package-Requires: ((emacs "27.1"))
 ;; Homepage: https://github.com/minad/osm
 
-;; This file is not part of GNU Emacs.
+;; This file is part of GNU Emacs.
 
 ;; This program is free software: you can redistribute it and/or modify
 ;; it under the terms of the GNU General Public License as published by
@@ -38,70 +38,79 @@
   :group 'web
   :prefix "osm-")
 
+(defvar osm--server-defaults
+  '(:min-zoom 2 :max-zoom 19 :max-connections 2 :subdomains ("a" "b" "c"))
+  "Default server properties.")
+
 (defcustom osm-server-list
   '((default
      :name "Mapnik"
      :description "Standard Mapnik map provided by OpenStreetMap"
-     :min-zoom 2 :max-zoom 19 :max-connections 2
-     :url "https://[abc].tile.openstreetmap.org/%z/%x/%y.png")
+     :url "https://%s.tile.openstreetmap.org/%z/%x/%y.png")
     (de
      :name "Mapnik(de)"
      :description "Localized Mapnik map provided by OpenStreetMap Germany"
-     :min-zoom 2 :max-zoom 19 :max-connections 2
-     :url "https://[abc].tile.openstreetmap.de/%z/%x/%y.png")
+     :url "https://%s.tile.openstreetmap.de/%z/%x/%y.png")
     (fr
      :name "Mapnik(fr)"
      :description "Localized Mapnik map by OpenStreetMap France"
-     :min-zoom 2 :max-zoom 19 :max-connections 2
-     :url "https://[abc].tile.openstreetmap.fr/osmfr/%z/%x/%y.png")
+     :url "https://%s.tile.openstreetmap.fr/osmfr/%z/%x/%y.png")
     (humanitarian
      :name "Humanitarian"
      :description "Humanitarian map provided by OpenStreetMap France"
-     :min-zoom 2 :max-zoom 19 :max-connections 2
-     :url "https://[abc].tile.openstreetmap.fr/hot/%z/%x/%y.png")
+     :url "https://%s.tile.openstreetmap.fr/hot/%z/%x/%y.png")
     (cyclosm
      :name "CyclOSM"
      :description "Bicycle-oriented map provided by OpenStreetMap France"
-     :min-zoom 2 :max-zoom 19 :max-connections 2
-     :url "https://[abc].tile.openstreetmap.fr/cyclosm/%z/%x/%y.png")
+     :url "https://%s.tile.openstreetmap.fr/cyclosm/%z/%x/%y.png")
     (openriverboatmap
      :name "OpenRiverBoatMap"
      :description "Waterways map provided by OpenStreetMap France"
-     :min-zoom 2 :max-zoom 19 :max-connections 2
-     :url "https://[abc].tile.openstreetmap.fr/openriverboatmap/%z/%x/%y.png")
+     :url "https://%s.tile.openstreetmap.fr/openriverboatmap/%z/%x/%y.png")
     (opentopomap
      :name "OpenTopoMap"
      :description "Topographical map provided by OpenTopoMap"
-     :min-zoom 2 :max-zoom 17 :max-connections 2
-     :url "https://[abc].tile.opentopomap.org/%z/%x/%y.png")
+     :url "https://%s.tile.opentopomap.org/%z/%x/%y.png")
     (opvn
-     :name "ÖPNV"
+     :name "ÖPNV" :max-zoom 18
      :description "Base layer with public transport information"
-     :min-zoom 2 :max-zoom 18 :max-connections 2
-     :url "http://[abc].tile.memomaps.de/tilegen/%z/%x/%y.png")
+     :url "http://%s.tile.memomaps.de/tilegen/%z/%x/%y.png")
     (stamen-watercolor
      :name "Stamen Watercolor"
      :description "Artistic map in watercolor style provided by Stamen"
-     :min-zoom 2 :max-zoom 19 :max-connections 2
-     :url "https://stamen-tiles-[abc].a.ssl.fastly.net/watercolor/%z/%x/%y.jpg")
+     :url "https://stamen-tiles-%s.a.ssl.fastly.net/watercolor/%z/%x/%y.jpg")
     (stamen-terrain
-     :name "Stamen Terrain"
+     :name "Stamen Terrain" :max-zoom 18
      :description "Map with hill shading provided by Stamen"
-     :min-zoom 2 :max-zoom 18 :max-connections 2
-     :url "https://stamen-tiles-[abc].a.ssl.fastly.net/terrain/%z/%x/%y.png")
+     :url "https://stamen-tiles-%s.a.ssl.fastly.net/terrain/%z/%x/%y.png")
     (stamen-toner-dark
      :name "Stamen Toner Dark"
      :description "Artistic map in toner style provided by Stamen"
-     :min-zoom 2 :max-zoom 19 :max-connections 2
-     :url "https://stamen-tiles-[abc].a.ssl.fastly.net/toner/%z/%x/%y.png")
+     :url "https://stamen-tiles-%s.a.ssl.fastly.net/toner/%z/%x/%y.png")
     (stamen-toner-light
      :name "Stamen Toner Lite"
      :description "Artistic map in toner style provided by Stamen"
-     :min-zoom 2 :max-zoom 19 :max-connections 2
-     :url "https://stamen-tiles-[abc].a.ssl.fastly.net/toner-lite/%z/%x/%y.png"))
+     :url "https://stamen-tiles-%s.a.ssl.fastly.net/toner-lite/%z/%x/%y.png"))
   "List of tile servers."
   :type '(alist :key-type symbol :value-type plist))
 
+(defcustom osm-pin-colors
+  '((selected-bookmark . "#e20")
+    (bookmark . "#f80")
+    (center . "#f08")
+    (org-link . "#7a9"))
+  "Colors of pins."
+  :type '(alist :key-type symbol :value-type string))
+
+(defcustom osm-home
+  (let ((lat (bound-and-true-p calendar-latitude))
+        (lon (bound-and-true-p calendar-longitude)))
+    (if (and lat lon)
+        (list lat lon 12)
+      (list 0 0 3)))
+  "Home coordinates, latitude, longitude and zoom level."
+  :type '(list number number number))
+
 (defcustom osm-large-step 256
   "Scroll step in pixel."
   :type 'integer)
@@ -130,15 +139,19 @@ Should be at least 7 days according to the server usage policies."
 
 (defvar osm-mode-map
   (let ((map (make-sparse-keymap)))
+    (define-key map [org-link] #'ignore)
+    (define-key map [center] #'ignore)
+    (define-key map [selected-bookmark] #'ignore)
+    (define-key map [bookmark mouse-1] #'osm-bookmark-select-click)
+    (define-key map [bookmark mouse-2] #'osm-bookmark-select-click)
+    (define-key map [bookmark mouse-3] #'osm-bookmark-select-click)
     (define-key map "+" #'osm-zoom-in)
     (define-key map "-" #'osm-zoom-out)
     (define-key map " " #'osm-zoom-in)
     (define-key map (kbd "S-SPC") #'osm-zoom-out)
-    (define-key map "\d" #'osm-zoom-out)
-    (define-key map [mouse-1] #'osm-zoom-click)
+    (define-key map [mouse-1] #'osm-center-click)
     (define-key map [mouse-2] #'osm-org-link-click)
     (define-key map [mouse-3] #'osm-bookmark-set-click)
-    (define-key map [S-mouse-3] #'osm-bookmark-delete-click)
     (define-key map [down-mouse-1] #'osm-drag)
     (define-key map [down-mouse-2] #'osm-drag)
     (define-key map [down-mouse-3] #'osm-drag)
@@ -154,6 +167,8 @@ Should be at least 7 days according to the server usage policies."
     (define-key map [M-down] #'osm-down-down)
     (define-key map [M-left] #'osm-left-left)
     (define-key map [M-right] #'osm-right-right)
+    (define-key map "d" #'osm-bookmark-delete)
+    (define-key map "\d" #'osm-bookmark-delete)
     (define-key map "c" #'clone-buffer)
     (define-key map "h" #'osm-home)
     (define-key map "g" #'osm-goto)
@@ -170,14 +185,21 @@ Should be at least 7 days according to the server usage policies."
   "Keymap used by `osm-mode'.")
 
 (defconst osm--placeholder
-  (list :type 'xbm :width 256 :height 256
-        :data (make-bool-vector (* 256 256) nil))
+  '(:type svg :width 256 :height 256
+    :data "<svg width='256' height='256' version='1.1' xmlns='http://www.w3.org/2000/svg'>
+  <defs>
+    <pattern id='grid' width='16' height='16'  patternUnits='userSpaceOnUse'>
+      <path d='m 0 0 l 0 16 16 0' fill='none' stroke='#888888'/>
+    </pattern>
+  </defs>
+  <rect width='256' height='256' fill='url(#grid)'/>
+</svg>")
   "Placeholder image for tiles.")
 
 (defvar osm--search-history nil
   "Minibuffer search history used by `osm-search'.")
 
-(defvar osm--clean-cache 0
+(defvar osm--purge-directory 0
   "Last time the tile cache was cleaned.")
 
 (defvar osm--tiles nil
@@ -186,8 +208,8 @@ Should be at least 7 days according to the server usage policies."
 (defvar osm--cookie 0
   "Tile cache cookie.")
 
-(defvar-local osm--url-index 0
-  "Current url index to query the servers in a round-robin fashion.")
+(defvar-local osm--subdomain-index 0
+  "Subdomain index to query the servers in a round-robin fashion.")
 
 (defvar-local osm--queue nil
   "Download queue of tiles.")
@@ -216,11 +238,11 @@ Should be at least 7 days according to the server usage policies."
 (defvar-local osm--y nil
   "X coordinate on the map in pixel.")
 
-(defvar-local osm--bookmark-positions
-  "Bookmark positions.")
+(defvar-local osm--pins nil
+  "Pin hash table.")
 
-(defvar-local osm--transient-pin nil
-  "Transient pin.")
+(defvar-local osm--transient-pins nil
+  "Transient pins.")
 
 (defun osm--boundingbox-to-zoom (lat1 lat2 lon1 lon2)
   "Compute zoom level from boundingbox LAT1 to LAT2 and LON1 to LON2."
@@ -266,32 +288,20 @@ Should be at least 7 days according to the server usage policies."
   "Convert LAT/ZOOM to y coordinate in pixel."
   (floor (* 256 (expt 2.0 zoom) (osm--lat-to-normalized-y lat))))
 
-(defun osm--home-coordinates ()
-  "Return home coordinate triple."
-  (let ((lat (bound-and-true-p calendar-latitude))
-        (lon (bound-and-true-p calendar-longitude))
-        (zoom 11))
-    (unless (and lat lon)
-      (setq lat 0 lon 0 zoom 2))
-    (list lat lon zoom)))
-
 (defun osm--server-property (prop)
   "Return server property PROP."
-  (plist-get (alist-get osm-server osm-server-list) prop))
+  (or (plist-get (alist-get osm-server osm-server-list) prop)
+      (plist-get osm--server-defaults prop)))
 
 (defun osm--tile-url (x y zoom)
   "Return tile url for coordinate X, Y and ZOOM."
-  (let ((url (osm--server-property :url))
-        (count 1))
-    (save-match-data
-      (when (string-match "\\`\\(.*\\)\\[\\(.*\\)\\]\\(.*\\)\\'" url)
-        (setq count (- (match-end 2) (match-beginning 2))
-              url (concat (match-string 1 url)
-                          (char-to-string (aref (match-string 2 url) osm--url-index))
-                          (match-string 3 url)))))
+  (let* ((url (osm--server-property :url))
+         (sub (osm--server-property :subdomains))
+         (count (length sub)))
     (prog1
-        (format-spec url `((?z . ,zoom) (?x . ,x) (?y . ,y)))
-      (setq osm--url-index (mod (1+ osm--url-index) count)))))
+        (format-spec url `((?z . ,zoom) (?x . ,x) (?y . ,y)
+                           (?s . ,(nth (mod osm--subdomain-index count) sub))))
+      (setq osm--subdomain-index (mod (1+ osm--subdomain-index) count)))))
 
 (defun osm--tile-file (x y zoom)
   "Return tile file name for coordinate X, Y and ZOOM."
@@ -314,10 +324,7 @@ Should be at least 7 days according to the server usage policies."
 (defun osm--download ()
   "Download next tile in queue."
   (when-let (job (and (< (length osm--active)
-                         (* (save-match-data
-                              (if (string-match "\\[\\(.*\\)\\]"
-                                                (osm--server-property :url))
-                                  (- (match-end 1) (match-beginning 1)) 1))
+                         (* (length (osm--server-property :subdomains))
                             (osm--server-property :max-connections)))
                       (pop osm--queue)))
     (push job osm--active)
@@ -374,14 +381,31 @@ Should be at least 7 days according to the server usage policies."
                        (lambda () (eq (car-safe last-input-event) 'mouse-movement))
                        (lambda () (setq track-mouse nil)))))
 
-(defun osm-zoom-click (event)
-  "Zoom to the location of the click EVENT."
+(defun osm--zoom-in-wheel (_n)
+  "Zoom in with the mouse wheel."
+  (pcase-let ((`(,x . ,y) (posn-x-y (event-start last-input-event))))
+    (when (< osm--zoom (osm--server-property :max-zoom))
+      (cl-incf osm--x (/ (- x osm--wx) 2))
+      (cl-incf osm--y (/ (- y osm--wy) 2))
+      (osm-zoom-in))))
+
+(defun osm--zoom-out-wheel (_n)
+  "Zoom out with the mouse wheel."
+  (pcase-let ((`(,x . ,y) (posn-x-y (event-start last-input-event))))
+    (when (> osm--zoom (osm--server-property :min-zoom))
+      (cl-decf osm--x (- x osm--wx))
+      (cl-decf osm--y (- y osm--wy))
+      (osm-zoom-out))))
+
+(defun osm-center-click (event)
+  "Center to the location of the click EVENT."
   (interactive "e")
   (pcase-let ((`(,x . ,y) (posn-x-y (event-start event))))
     (when (< osm--zoom (osm--server-property :max-zoom))
       (cl-incf osm--x (- x osm--wx))
       (cl-incf osm--y (- y osm--wy))
-      (osm-zoom-in))))
+      (osm--put-transient-pin 'center osm--x osm--y "Center")
+      (osm--update))))
 
 (defun osm-bookmark-set-click (event)
   "Create bookmark at position of click EVENT."
@@ -391,21 +415,26 @@ Should be at least 7 days according to the server usage policies."
                (osm--y (+ osm--y (- y osm--wy))))
     (osm-bookmark-set)))
 
-(defun osm-bookmark-delete-click (event)
-  "Delete bookmark at position of click EVENT."
+(defun osm-bookmark-select-click (event)
+  "Select bookmark at position of click EVENT."
   (interactive "@e")
   (pcase-let* ((`(,x . ,y) (posn-x-y (event-start event)))
                (x (+ osm--x (- x osm--wx)))
                (y (+ osm--y (- y osm--wy)))
                (min most-positive-fixnum)
                (found nil))
-    (cl-loop for (p q . name) in osm--bookmark-positions
-             for d = (+ (* (- p x) (- p x)) (* (- q y) (- q y)))
-             if (and (< d 2500) (< d min)) do
-             (setq min d found name))
-    (unless found
-      (error "No bookmark at point"))
-    (osm-bookmark-delete found)))
+    (dolist (bm bookmark-alist)
+      (when (eq (bookmark-prop-get bm 'handler) #'osm-bookmark-jump)
+        (let* ((coord (bookmark-prop-get bm 'coordinates))
+               (p (osm--lon-to-x (cadr coord) osm--zoom))
+               (q (osm--lat-to-y (car coord) osm--zoom))
+               (d (+ (* (- p x) (- p x)) (* (- q y) (- q y)))))
+          (when (and (>= q y) (< q (+ y 50)) (>= p (- x 20)) (< p (+ x 20)) (< d min))
+            (setq min d found (list p q (car bm)))))))
+    (when found
+      (message "Selected '%s'" (cddr found))
+      (apply #'osm--put-transient-pin 'selected-bookmark found)
+      (osm--update))))
 
 (defun osm-org-link-click (event)
   "Store link at position of click EVENT."
@@ -413,7 +442,9 @@ Should be at least 7 days according to the server usage policies."
   (pcase-let* ((`(,x . ,y) (posn-x-y (event-start event)))
                (osm--x (+ osm--x (- x osm--wx)))
                (osm--y (+ osm--y (- y osm--wy))))
-    (call-interactively 'org-store-link)))
+    (call-interactively 'org-store-link)
+    (osm--put-transient-pin 'org-link osm--x osm--y "Org Link"))
+  (osm--update))
 
 (defun osm-zoom-in (&optional n)
   "Zoom N times into the map."
@@ -485,11 +516,11 @@ Should be at least 7 days according to the server usage policies."
   (interactive "p")
   (osm-right-right (- (or n 1))))
 
-(defun osm--clean-cache ()
-  "Clean tile cache."
+(defun osm--purge-directory ()
+  "Clean tile directory."
   (when (and (integerp osm-max-age)
-             (> (- (float-time) osm--clean-cache) (* 60 60 24)))
-    (setq osm--clean-cache (float-time))
+             (> (- (float-time) osm--purge-directory) (* 60 60 24)))
+    (setq osm--purge-directory (float-time))
     (run-with-idle-timer
      30 nil
      (lambda ()
@@ -511,7 +542,6 @@ Should be at least 7 days according to the server usage policies."
   (dolist (type '(svg jpeg png))
     (unless (image-type-available-p type)
       (warn "osm: Support for %s images is missing" type)))
-  (osm--clean-cache)
   (setq-local osm-server osm-server
               line-spacing nil
               cursor-type nil
@@ -526,43 +556,47 @@ Should be at least 7 days according to the server usage policies."
               buffer-read-only t
               fringe-indicator-alist '((truncation . nil))
               revert-buffer-function #'osm--revert
-              mwheel-scroll-up-function #'osm-down
-              mwheel-scroll-down-function #'osm-up
-              mwheel-scroll-left-function #'osm-left
-              mwheel-scroll-right-function #'osm-right
+              mouse-wheel-progressive-speed nil
+              mwheel-scroll-up-function #'osm--zoom-out-wheel
+              mwheel-scroll-down-function #'osm--zoom-in-wheel
+              mwheel-scroll-left-function #'osm--zoom-out-wheel
+              mwheel-scroll-right-function #'osm--zoom-in-wheel
               bookmark-make-record-function #'osm--make-bookmark)
   (add-hook 'window-size-change-functions #'osm--resize nil 'local))
 
-(defun osm--bookmark-positions ()
-  "Compute bookmark positions."
-  (bookmark-maybe-load-default-file)
-  (setq osm--bookmark-positions
-        (cl-loop
-         for bm in bookmark-alist
-         if (eq (bookmark-prop-get bm 'handler) #'osm-bookmark-jump)
-         collect
-         (let* ((coord (bookmark-prop-get bm 'coordinates))
-                (px (osm--lon-to-x (cadr coord) osm--zoom))
-                (py (osm--lat-to-y (car coord) osm--zoom)))
-           `(,px ,py . ,(car bm))))))
-
-(defun osm--pin-position (x y p q)
-  "Return position of pin P/Q in tile X/Y."
+(defun osm--pin-at-p (x y p q)
+  "Return non-nil if pin P/Q is inside tile X/Y."
   (setq x (* x 256) y (* y 256))
-  (and (>= p (- x 100)) (< p (+ x 256 100))
-       (>= q (- y 100)) (< q (+ y 256 100))
-       (cons (- p x) (- q y))))
+  (and (>= p (- x 32)) (< p (+ x 256 32))
+       (>= q y) (< q (+ y 256 64))))
 
-(defun osm--pins-at (x y)
-  "Get pins at X/Y."
-  ;; TODO Optimized k2 tree?
-  (let ((pins (cl-loop for (p q . _name) in osm--bookmark-positions
-                       for pin = (osm--pin-position x y p q)
-                       if pin collect (cons "#ff8800" pin))))
-    (if-let (pin (and osm--transient-pin
-                      (osm--pin-position x y osm--x osm--y)))
-        (nconc pins (list (cons "#ff0088" pin)))
-      pins)))
+(defun osm--put-pin (id x y help)
+  "Put pin at X/Y with HELP and ID in pins hash table."
+  (let ((x0 (/ x 256)) (y0 (/ y 256)))
+    (push `(,(- x (* x0 256)) ,(- y (* y0 256)) ,id . ,help)
+          (gethash (cons x0 y0) osm--pins))
+    (cl-loop
+     for i from -1 to 1 do
+     (cl-loop
+      for j from -1 to 0 do
+      (let ((x1 (/ (+ x (* 32 i)) 256))
+            (y1 (/ (+ y (* 64 j)) 256)))
+        (unless (and (= x0 x1) (= y0 y1))
+          (push `(,(- x (* x1 256)) ,(- y (* y1 256)) ,id . ,help)
+                (gethash (cons x1 y1) osm--pins))))))))
+
+(defun osm--update-pins ()
+  "Compute pin positions."
+  (setq osm--pins (make-hash-table :test #'equal))
+  (dolist (pin osm--transient-pins)
+    (apply #'osm--put-pin pin))
+  (bookmark-maybe-load-default-file)
+  (dolist (bm bookmark-alist)
+    (when (eq (bookmark-prop-get bm 'handler) #'osm-bookmark-jump)
+      (let* ((coord (bookmark-prop-get bm 'coordinates))
+             (x (osm--lon-to-x (cadr coord) osm--zoom))
+             (y (osm--lat-to-y (car coord) osm--zoom)))
+        (osm--put-pin 'bookmark x y (car bm))))))
 
 (autoload 'svg--image-data "svg")
 (defun osm--make-tile (x y)
@@ -571,30 +605,38 @@ Should be at least 7 days according to the server usage policies."
     (when (file-exists-p file)
       `(image
         :width 256 :height 256
-        ,@(if-let (pins (osm--pins-at x y))
-              (list :type 'svg :base-uri file
-                    :data (concat "<svg width='256' height='256' version='1.1'
+        ,@(if-let (pins (gethash (cons x y) osm--pins))
+              (let* ((areas nil)
+                     (svg-pins
+                     (mapconcat
+                      (pcase-lambda (`(,p ,q ,id . ,help))
+                        (push `((poly . [,p ,q ,(- p 20) ,(- q 40) ,p ,(- q 50) ,(+ p 20) ,(- q 40) ])
+                                ,id (help-echo ,(truncate-string-to-width help 20 0 nil t) pointer hand))
+                              areas)
+                        ;; https://commons.wikimedia.org/wiki/File:Simpleicons_Places_map-marker-1.svg
+                        (format "
+<g fill='%s' stroke='#000000' stroke-width='9' transform='translate(%s %s) scale(0.09) translate(-256 -512)'>
+<path d='M256 0C167.641 0 96 71.625 96 160c0 24.75 5.625 48.219 15.672
+69.125C112.234 230.313 256 512 256 512l142.594-279.375
+C409.719 210.844 416 186.156 416 160C416 71.625 344.375
+0 256 0z M256 256c-53.016 0-96-43-96-96s42.984-96 96-96
+c53 0 96 43 96 96S309 256 256 256z'/>
+</g>" (alist-get id osm-pin-colors) p q))
+                      pins "")))
+                (list :type 'svg :base-uri file :map areas
+                      :data (concat "<svg width='256' height='256' version='1.1'
 xmlns='http://www.w3.org/2000/svg' xmlns:xlink='http://www.w3.org/1999/xlink'>
 <image xlink:href='"
-                                  (if (> emacs-major-version 27)
-                                      (file-name-nondirectory file)
-                                    ;; NOTE: On Emacs 27, :base-uri and embedding by file
-                                    ;; path is not supported. Use the less efficient base64 encoding.
-                                    (svg--image-data
-                                     file
-                                     (if (member (file-name-extension file) '("jpg" "jpeg"))
-                                         "image/jpeg" "image/png")
-                                     nil))
-                                  "' height='256' width='256'/>"
-                                  (mapconcat
-                                   (pcase-lambda (`(,color ,x . ,y))
-                                     (format "
-<g fill='%s' stroke='#000000' transform='translate(%s %s)'>
-<polygon points='0 0 7 -35 -7 -35'/>
-<circle cx='0' cy='-35' r='12'/>
-</g>" color x y))
-                                   pins "")
-                                  "</svg>"))
+                                    (if (> emacs-major-version 27)
+                                        (file-name-nondirectory file)
+                                      ;; NOTE: On Emacs 27, :base-uri and embedding by file
+                                      ;; path is not supported. Use the less efficient base64 encoding.
+                                      (svg--image-data
+                                       file
+                                       (if (member (file-name-extension file) '("jpg" "jpeg"))
+                                           "image/jpeg" "image/png")
+                                       nil))
+                                    "' height='256' width='256'/>" svg-pins "</svg>")))
             (list :type
                   (if (member (file-name-extension file) '("jpg" "jpeg"))
                       'jpeg 'png)
@@ -602,7 +644,8 @@ xmlns='http://www.w3.org/2000/svg' xmlns:xlink='http://www.w3.org/1999/xlink'>
 
 (defun osm--get-tile (x y)
   "Get tile at X/Y."
-  (if (and osm--transient-pin (osm--pin-position x y osm--x osm--y))
+  (if (cl-loop for (_id p q . _rest) in osm--transient-pins
+               thereis (osm--pin-at-p x y p q))
       (osm--make-tile x y)
     (let* ((key `(,osm-server ,osm--zoom ,x . ,y))
            (tile (and osm--tiles (gethash key osm--tiles))))
@@ -637,7 +680,7 @@ xmlns='http://www.w3.org/2000/svg' xmlns:xlink='http://www.w3.org/1999/xlink'>
 (defun osm-home ()
   "Go to home coordinates."
   (interactive)
-  (osm--goto (osm--home-coordinates) nil))
+  (osm--goto osm-home nil))
 
 (defun osm--queue-info ()
   "Return queue info string."
@@ -659,15 +702,15 @@ xmlns='http://www.w3.org/2000/svg' xmlns:xlink='http://www.w3.org/1999/xlink'>
   (when (eq major-mode #'osm-mode)
     (osm--update)))
 
-(defun osm--header ()
+(defun osm--update-header ()
   "Update header line."
   (let* ((meter-per-pixel (/ (* 156543.03 (cos (/ (osm--lat) (/ 180.0 float-pi)))) (expt 2 osm--zoom)))
-         (meter '(1 5 10 50 100 500 1000 5000 10000 50000 100000 500000 1000000 5000000 10000000))
          (server (osm--server-property :name))
-         (idx 0))
-    (while (and (< idx (1- (length meter))) (< (/ (nth (1+ idx) meter) meter-per-pixel) 100))
+         (meter 1) (idx 0)
+         (factor '(2 2.5 2)))
+    (while (and (< idx 20) (< (/ (* meter (nth (mod idx 3) factor)) meter-per-pixel) 150))
+      (setq meter (round (* meter (nth (mod idx 3) factor))))
       (cl-incf idx))
-    (setq meter (nth idx meter))
     (setq-local
      header-line-format
      (list
@@ -693,41 +736,66 @@ xmlns='http://www.w3.org/2000/svg' xmlns:xlink='http://www.w3.org/1999/xlink'>
   (unless (eq major-mode #'osm-mode)
     (error "Not an osm-mode buffer"))
   (rename-buffer (osm--buffer-name) 'unique)
-  (osm--header)
-  (osm--bookmark-positions)
+  (osm--update-sizes)
+  (osm--update-header)
+  (osm--update-pins)
+  (osm--update-buffer)
+  (osm--process-queue)
+  (osm--purge-tiles)
+  (osm--purge-directory))
+
+(defun osm--update-sizes ()
+  "Update window sizes."
   (let* ((windows (or (get-buffer-window-list) (list (frame-root-window))))
          (win-width (cl-loop for w in windows maximize (window-pixel-width w)))
          (win-height (cl-loop for w in windows maximize (window-pixel-height w))))
     (setq osm--wx (/ win-width 2)
           osm--wy (/ win-height 2)
           osm--nx (1+ (ceiling win-width 256))
-          osm--ny (1+ (ceiling win-height 256)))
-    (with-silent-modifications
-      (erase-buffer)
-      (dotimes (_j osm--ny)
-        (insert (make-string osm--nx ?\s) "\n"))
-      (goto-char (point-min))
-      (dotimes (j osm--ny)
-        (dotimes (i osm--nx)
-          (let* ((x (+ i (/ (- osm--x osm--wx) 256)))
-                 (y (+ j (/ (- osm--y osm--wy) 256)))
-                 (tile (osm--get-tile x y)))
-            (osm--display-tile x y tile)
-            (unless tile (osm--enqueue x y))))))
-    (cl-incf osm--cookie)
-    (when (and osm--tiles (> (hash-table-count osm--tiles) osm-max-tiles))
-      (let (items)
-        (maphash (lambda (k v) (push (cons (car v) k) items)) osm--tiles)
-        (setq items (sort items #'car-less-than-car))
-        (dotimes (_ (- (hash-table-count osm--tiles) osm-max-tiles))
-          (remhash (cdr (pop items)) osm--tiles))))
-    (setq osm--queue
-          (sort osm--queue
-                (pcase-lambda (`(,x1 ,y1 . ,_z1) `(,x2 ,y2 . ,_z2))
-                  (setq x1 (- x1 (/ osm--x 256)) y1 (- y1 (/ osm--y 256))
-                        x2 (- x2 (/ osm--x 256)) y2 (- y2 (/ osm--y 256)))
-                  (< (+ (* x1 x1) (* y1 y1)) (+ (* x2 x2) (* y2 y2))))))
-    (osm--download)))
+          osm--ny (1+ (ceiling win-height 256)))))
+
+(defun osm--update-buffer ()
+  "Update buffer display."
+  (with-silent-modifications
+    (erase-buffer)
+    (dotimes (_j osm--ny)
+      (insert (make-string osm--nx ?\s) "\n"))
+    (goto-char (point-min))
+    (dotimes (j osm--ny)
+      (dotimes (i osm--nx)
+        (let* ((x (+ i (/ (- osm--x osm--wx) 256)))
+               (y (+ j (/ (- osm--y osm--wy) 256)))
+               (tile (osm--get-tile x y)))
+          (osm--display-tile x y tile)
+          (unless tile (osm--enqueue x y)))))))
+
+(defun osm--process-queue ()
+  "Process the download queue."
+  (setq osm--queue
+        (sort
+         (cl-loop for job in osm--queue
+                  for (x y . zoom) = job
+                  for i = (- x (/ (- osm--x osm--wx) 256))
+                  for j = (- y (/ (- osm--y osm--wy) 256))
+                  if (and (= zoom osm--zoom)
+                          (>= i 0) (< i osm--nx)
+                          (>= j 0) (< j osm--ny))
+                  collect job)
+         (pcase-lambda (`(,x1 ,y1 . ,_z1) `(,x2 ,y2 . ,_z2))
+           (setq x1 (- x1 (/ osm--x 256)) y1 (- y1 (/ osm--y 256))
+                 x2 (- x2 (/ osm--x 256)) y2 (- y2 (/ osm--y 256)))
+           (< (+ (* x1 x1) (* y1 y1)) (+ (* x2 x2) (* y2 y2))))))
+  (osm--download))
+
+(defun osm--purge-tiles ()
+  "Purge old tiles from the tile cache."
+  (cl-incf osm--cookie)
+  (when (and osm--tiles (> (hash-table-count osm--tiles) osm-max-tiles))
+    (let (items)
+      (maphash (lambda (k v) (push (cons (car v) k) items)) osm--tiles)
+      (setq items (sort items #'car-less-than-car))
+      (dotimes (_ (- (hash-table-count osm--tiles) osm-max-tiles))
+        (remhash (cdr (pop items)) osm--tiles)))))
 
 (defun osm--make-bookmark ()
   "Make osm bookmark record."
@@ -767,7 +835,7 @@ xmlns='http://www.w3.org/2000/svg' xmlns:xlink='http://www.w3.org/1999/xlink'>
   (with-current-buffer
       (or
        (and (eq major-mode #'osm-mode) (current-buffer))
-       (pcase-let* ((`(,def-lat ,def-lon ,def-zoom) (or at (osm--home-coordinates)))
+       (pcase-let* ((`(,def-lat ,def-lon ,def-zoom) (or at osm-home))
                     (def-x (osm--lon-to-x def-lon def-zoom))
                     (def-y (osm--lat-to-y def-lat def-zoom))
                     (def-server (or server osm-server)))
@@ -788,21 +856,31 @@ xmlns='http://www.w3.org/2000/svg' xmlns:xlink='http://www.w3.org/1999/xlink'>
             osm--active nil
             osm--queue nil))
     (when (or (not (and osm--x osm--y)) at)
-      (let ((buffer (current-buffer))
-            (sym (make-symbol "osm--remove-transient-pin")))
-        (fset sym (lambda ()
-                    (with-current-buffer buffer
-                      (setq osm--transient-pin nil)
-                      (remove-hook 'pre-command-hook sym))))
-        (add-hook 'pre-command-hook sym))
-      (setq at (or at (osm--home-coordinates))
+      (setq at (or at osm-home)
             osm--zoom (nth 2 at)
             osm--x (osm--lon-to-x (nth 1 at) osm--zoom)
-            osm--y (osm--lat-to-y (nth 0 at) osm--zoom)
-            osm--transient-pin t))
+            osm--y (osm--lat-to-y (nth 0 at) osm--zoom))
+      (osm--put-transient-pin 'center osm--x osm--y "Center"))
     (prog1 (pop-to-buffer (current-buffer))
       (osm--update))))
 
+(defun osm--put-transient-pin (id x y help)
+  "Set transient pin at X/Y with ID and HELP."
+  (let ((buffer (current-buffer))
+        (sym (make-symbol "osm--remove-transient-pin")))
+    (fset sym (lambda ()
+                (with-current-buffer buffer
+                  (remove-hook 'pre-command-hook sym)
+                  (setq osm--transient-pins (assq-delete-all id osm--transient-pins))
+                  ;; HACK: handle bookmark deletion
+                  (when (and (eq this-command #'osm-bookmark-delete)
+                             (eq id 'selected-bookmark))
+                    (osm-bookmark-delete help)
+                    (setq this-command #'ignore)))))
+    (add-hook 'pre-command-hook sym)
+    (setf (alist-get id osm--transient-pins)
+          (list x y help))))
+
 ;;;###autoload
 (defun osm-goto (lat lon zoom)
   "Go to LAT/LON/ZOOM."
@@ -939,8 +1017,8 @@ MSG is a message prefix string."
                    #'osm-up-up #'osm-down-down #'osm-left-left #'osm-right-right
                    #'osm-zoom-out #'osm-zoom-in #'osm-bookmark-set))
   (put sym 'command-modes '(osm-mode)))
-(dolist (sym (list #'osm-drag #'osm-zoom-click #'osm-org-link-click
-                    #'osm-bookmark-set-click #'osm-bookmark-delete-click))
+(dolist (sym (list #'osm-drag #'osm-center-click #'osm-org-link-click
+                    #'osm-bookmark-set-click #'osm-bookmark-select-click))
   (put sym 'completion-predicate #'ignore))
 
 (provide 'osm)
